;
        TITLE   SFI_SHORTEST_PATH_FINDER
        PAGE    198,81
;
; STARFLIGHT I Shortest Path Finder V 1.1
; (C)opyright 1989-2000, David R. Sheldon
;
;   This program accepts two X,Y points on the
; command line and displays the shortest travel
; path. In the Starflight I game, traveling from
; 0,0 to 0,5 uses the same amount of fuel as
; traveling from 0,0 to 5,5. Although the later
; may be a greater distance, the two are treated
; the same. This does not follow normal two-
; dimensional math. (That's probably why the
; game has so much going on in the constellation
; Pythagoras; 3 habitable planets, 2 artifacts,
; and Ben and Anna's conundrum.) To further
; complicate matters, continuum flux are present
; that allow travel from one point to another
; without use of fuel. Mathematically, this has
; the effect of placing discontinuities or
; "folds" all over the two-dimensional plane.
; Finding the shortest path between two points
; can be a considerable task and I always find
; myself wondering if my answer is right. This
; program solves the problem. For a long time, I
; used the 248,1-55,99 flux to go from 234,20 to
; 55,32. It may well be the simplest path but
; the program found one 5 sectors shorter.
;   This code may be adapted for use in the map
; generator/viewer program as an optional
; caliper mode (cool). That program will only
; be available to registered users. I would like
; a program to find the shortest distance to
; cover a number of systems, like a mission to
; colonize planets. The combinatorial math goes
; nuts after about 16 systems or so. There are
; just too many combinations to deal with. The
; closest I could get would be a program that
; goes to one system, finds the closest system
; in a list, and so on. I may write that one.
;   Except for bubble sort (widely published),
; this is the first time I've had a good
; application for a recursive routine (i.e. one
; that calls itself). It is used to continuously
; test multi-flux paths. Each time a flux is
; added to the list, it increases the search
; depth, and the routine calls itself to further
; test for more flux. Simple, eh.
;   I was a little disappointed with the speed
; of the initial code, so I changed some of it
; to make it faster. Naturally, the longer the
; direct path distance, the more possible flux
; combinations there are to test, and the longer
; it takes to find the path. I tried optimizing
; the flux list order prior to the search based
; on the origin and destination values. It only
; provided about a 1.5% improvement using a set
; of 20 standard input pairs. I then reorganized
; the flux list so that longer flux are tested
; first. It was 25% faster! Well, of course; The
; "standard" pairs are mostly from map corner to
; map corner. It does make sense to test longer
; flux first, though. If the journey is of any
; length, it is likely to use one or more of the
; longer flux. I don't really know that it's
; optimal, but I prefer to speed up longer path
; searches, perhaps at the expense of shorter
; ones. There are 3,025,000,000 possible pairs.
; I don't know how fast your machine is, but it
; would take a while to test them all on mine.
; To do it in a day, 35,012 pairs would have to
; be tested every second. I wish! I'm sure that
; most of the engineers at Intel are working on
; 700 gigahertz. (Pentium MCMLXXXIV)
;   Version 1.1 was a recent update, where I
; added the title and usage messages for public
; release, and explained things a little better
; in the remarks. I will allow distribution of
; this source file, as it doesn't reveal any
; floating-point math, graphics, disk buffer, or
; other guarded routines. Some users are
; programmers, as well, and may want to adapt
; the code to suit their own personal use. For
; example, converting this program for use with
; STARFLIGHT II would be easy. (When I know
; where all the flux are, I'll do so.) If you
; want to do it, change the "program constant"
; equates at the beginning of the program text
; and list the flux in CFLUXS (near the end).
;
; Use:
; >SF1SPF11 Xo,Yo Xd,Yd
; where:
; Xo = origin X
; Yo = origin Y
; Xd = destination X
; Yd = destination Y
;
; Example:
; >SF1SPF11 125,100 132,165
;
; Notes:
; 1) valid X values range from 0 to 249
; 2) valid Y values range from 0 to 219
; 3) no penalty imposed for 45 degree travel
; 4) non-numeric characters are skipped,
;    commas are optional (may be spaces)
; 5) input beyond 4 values is ignored
; 6) usage is displayed for invalid input
;
;-----------------------------------------------
;
; program constants
;
; MAXIMX < 256, MAXIMY < 256, FLXQTY < 128
;
MAXIMX  EQU     249           ;maximum X value
MAXIMY  EQU     219           ;maximum Y value
FLXQTY  EQU     32            ;total game flux
;
;-----------------------------------------------
;
; uninitialized data area sizes
;
SZPFLP  EQU     2*FLXQTY      ;PFLPTR = 32 words
SZCBPA  EQU     4*FLXQTY+4    ;CBPATH = 66 words
SZSPLI  EQU     2*FLXQTY+2    ;SPLIST = 33 words
;
; uninitialized data offsets relative to BUFFER
;
CBPATH  EQU     SZPFLP        ;current best path
SPLIST  EQU     CBPATH+SZCBPA ;search path list
PFLIST  EQU     SPLIST+SZSPLI ;prospective flux
;
;-----------------------------------------------
;
; abrieviations
;
BPT     EQU     BYTE PTR
WPT     EQU     WORD PTR
OFS     EQU     OFFSET
SHO     EQU     SHORT
;
;-----------------------------------------------
;
_TEXT   SEGMENT WORD PUBLIC 'CODE'
        ASSUME  CS:_TEXT,DS:_TEXT
;
; DOS command line
;
        ORG     80h
DCLINE  LABEL   BYTE
;
; program offsets above PSP
;
        ORG     100h
;
;-----------------------------------------------
;
_main   PROC    NEAR
;
; parse command line
;
        MOV     SI,OFS DCLINE
        LODSB
        CBW
        XCHG    AX,CX
        JCXZ    S1SPF0
;
        MOV     DL,0Ah        ;divisor
        CALL    PARSE         ;parse numeric
        JBE     S1SPF0
;
        XCHG    AX,BX
        CALL    PARSE         ;parse numeric
        JBE     S1SPF0
;
        MOV     BH,AL
        CALL    PARSE         ;parse numeric
        JBE     S1SPF0
;
        XCHG    AX,DI
        CALL    PARSE         ;parse numeric
        JC      S1SPF0
;
        XCHG    AL,AH
        OR      AX,DI
        CMP     AH,MAXIMY     ;AH = dest Y
        JA      S1SPF0
;
        CMP     BH,MAXIMY     ;BH = origin Y
        JA      S1SPF0
;
        CMP     AL,MAXIMX     ;AL = dest X
        JA      S1SPF0
;
        CMP     BL,MAXIMX     ;BL = origin X
        JBE     S1SPF1        ;all values good
;
; display usage message
;
;   If the command line is invalid, the title
; and usage messages are displayed using
; interrupt 21h, function 9 so that all output
; may be redirected.
;
S1SPF0: MOV     DX,OFS MTITLE
        MOV     AH,9
        INT     21h
;
; terminate with batch error flag
;
        MOV     AX,4C01h
        INT     21h
;
; display title
;
;   If the command line is valid, the title
; message is displayed using interrupt 29h so
; that its output will not be redirected. The
; travel and flux outputs may be redirected as
; interrupt 21h, functions 2 and 9 are used.
;   Although I believe most (if not all) of the
; registers are preserved across interrupt 29h
; calls, better safe than sorry; I push and pop.
;
S1SPF1: MOV     SI,OFS MTITLE
        PUSH    AX
        MOV     CX,OFS MUSAGE-OFS MTITLE
        PUSH    BX
;
S1SPF2: PUSH    CX
        LODSB
        PUSH    SI
        INT     29h
        POP     SI
        POP     CX
        LOOP    S1SPF2
;
; initialize current best path variables
;
        POP     BX
        MOV     BUFFER+CBPATH,BX ;origin
        POP     AX
        MOV     ULDEST-2,AX   ;ultimate dest
        CALL    CDIST         ;calculate dist
        MOV     CH,DL
        MOV     CBDIST-1,DL   ;best distance
        JCXZ    S1SPF8
;
; attempt to find a shorter path
;
        MOV BPT BUFFER+SPLIST+1,CL ;(CL = 0)
        MOV     BP,CX
        CALL    SPATH         ;SPATH recursive
;
; report results
;
;   The first point in the CBPATH list is always
; the users origin. Flux are initially listed as
; termination numbers. They are converted, here,
; into two end points and the destination is
; added to the end of the list. The origin may
; be a flux end point, and that flux may or may
; not be a part of the best path. The second
; point is always either the destination or a
; flux start point. If the first two points are
; the same, we know that the first reported item
; is a flux. (Normally, the first reported item
; will be a travel.) When we get to the end of
; the list, a similar test must be made to see
; if the last two points are the same. After
; conversion, the total number of points in the
; list is found by: N = 2 x (flux quantity) + 2
;
S1SPF3: CALL    DCRLF         ;car ret/line feed
        MOV     DI,OFS BUFFER+CBPATH+2
        MOV     AL,CBFLUX-1   ;best flux quantity
        CBW
        MOV     CX,AX
        MOV     SI,AX
        DEC     SI
        SHL     SI,1
        ADD     SI,DI
        SHL     AX,1
        SHL     AX,1
        ADD     DI,AX
        MOV     AX,ULDEST-2
        STD
        STOSW
        JCXZ    S1SPF5
;
        PUSH    CX
;
S1SPF4: LODSB
        CALL    TFLUX         ;get X,Y point
        XCHG    AX,BX
        STOSW
        DEC     SI
        XCHG    AX,BX
        CALL    FLOPP         ;opposite flux end
        CALL    TFLUX         ;get X,Y point
        XCHG    AX,BX
        STOSW
        LOOP    S1SPF4
;
        POP     CX
;
S1SPF5: MOV     SI,OFS BUFFER+CBPATH
        CLD
        MOV     DX,0A00h
        SHL     CX,1
        JZ      S1SPF7
;
        MOV     AX,[SI]
        CMP     AX,[SI+2]
        JNZ     S1SPF6
;
        INC     SI
        INC     SI
        INC     DX
        DEC     CX
;
S1SPF6: PUSH    CX
        CALL    DSPLY         ;display line item
        POP     CX
        LOOP    S1SPF6
;
        MOV     AX,[SI]
        CMP     AX,[SI+2]
        JZ      S1SPF8
;
S1SPF7: CALL    DSPLY         ;display line item
;
S1SPF8: CALL    DCRLF         ;car ret/line feed
        MOV     DX,OFS TRDIST
        MOV     AH,9
        INT     21h
        MOV     AL,CBDIST-1   ;best distance
        MOV     DH,0Ah        ;divisor
        CALL    DSVAL         ;dsiplay value
        CALL    DCRLF         ;car ret/line feed
;
; terminate without batch error flag
;
        MOV     AX,4C00h
        INT     21h
;
_main   ENDP
;
;-----------------------------------------------
;
SPATH   PROC    NEAR
;
; Find shortest path.
;
;   This routine is recursive. It uses most of
; the other routines and itself to find the
; shortest path from point A to point B. It
; works by making an initial flux list and, for
; each flux in that list, searching for further
; (deeper levels) flux.
;   In a more complex recursive routine, I would
; have used the BP register to establish a stack
; frame for variable storage. Not doing so frees
; that register for other use and reduces memory
; accesses (I think), which is faster. It also
; seems to simplify the code. The BP register
; holds 2 variables throughout execution. The
; upper 8 bits contain a value equaling the
; current best path distance, minus the distance
; consumed in the current search path list. The
; lower 8 bits contain the current search depth,
; which is also the number of flux in the
; current search path. This later value is used
; for a variety of tasks, such as calculating
; offsets into PFLPTR and SPLIST. Each time
; SPATH is called, this value is incremented.
; When the routine exits, it is decremented.
;   There are 3 main parts to the routine; (1)
; ID flux and look for a new best path, (2)
; recursion loop to test for next level flux,
; and (3) a section of code used by (1) to
; update the current best path variables when
; one is found. The first 2 parts are loops
; where a majority of the time is spent, and
; were written with a priority on speed.
;   When I wrote this program, I wrote all of
; the other routines first, leaving this one for
; last. It was surprisingly easy to write once I
; had the concept fixed in my head. It was even
; more of a surprise to see it work the first
; time. That happens about as often as the best
; Starflight path is a straight line.
;   Since then, I have spent some time massaging
; the code to speed it up. I tried one version
; where the recursion was placed inside the
; first loop (no second loop). It eliminated the
; need for a prospective flux list and the
; associated pointer list. It really looked
; great, as far as being cleaner. It was very
; slow, though. I figured out that it was
; testing deeper level paths before the shallow
; ones. This meant that it would test the paths
; with, say, 16 and greater flux before it
; tested the paths with 3 or 4 flux. This left
; the current best distance value large until
; the end of the search, and made for more
; tested paths. This is due to the way the flux
; are arranged on the map. Most best paths use 0
; to 7 flux, not 20.
;   I'm still trying to speed this thing up. If
; it is to be used in conjunction with the map
; generator/viewer program caliper, it is going
; to have to be a little faster on long paths.
;
;Call With:
; BL = most recent origin X
; BH = most recent origin Y
; BPl = search level
; BPh = (current best dist) - (search path dist)
; ULDEST = ultimate destination
; SPLIST = search path list
; PFLPTR = prospective flux pointer list
; PFLIST = prospective flux list(s)
; CBDIST \
; CBFLUX  > current best path
; CBPATH /
;
;Returns:
; BPl = search level
; SPLIST = search path list
; PFLPTR = prospective flux pointer list
; PFLIST = prospective flux list(s)
; CBDIST \
; CBFLUX  > current best path
; CBPATH /
;
;- - - - - - - -
;
; identify flux within range and better path
;
;   Initially, there were two seperate loops;
; one to create the flux list, and one to check
; for a new best path. Combining the two really
; provided a major speed improvement. Not only
; is there only one loop, but as the current
; best path gets shorter and shorter, fewer flux
; qualify for the flux list. Reducing the size
; of the flux list means fewer and shallower
; searches.
;
;- - - - - - - -
;
        MOV     AX,BP
        CBW
        SHL     AX,1
        XCHG    AX,SI
        MOV     DI,PFLPTR[SI]
        INC     BP            ;next search level
        MOV     SI,OFS CFLUXS ;flux list offset
        PUSH    DI
        MOV     CX,2*FLXQTY   ;loop double count
;
; find closest flux end
;
SPATH0: CALL    CDLAX         ;calculate dist
        MOV     DH,DL
        DEC     CX            ;count twice
        CALL    CDLAX         ;calculate dist
        MOV     AX,BP
        MOV     AL,CL
        CMP     DL,DH
        JA      SPATH1
;
        JZ      SPATH6        ;both ends same
;
        MOV     DH,DL
        INC     AX
;
; test to see if it's a prospect
;
SPATH1: CMP     AH,DH
        JB      SPATH6
;
; make sure it's not already in the search path
; (avoids doubling back on same flux - critical)
;
        MOV     DL,AL
        PUSH    CX
        MOV     CX,BP
        PUSH    SI
        MOV     CH,0
        DEC     CX
        JZ      SPATH3        ;null search path
;
        CALL    FLOPP         ;opposite flux end
        MOV     AH,AL
        MOV     SI,OFS BUFFER+SPLIST+2
;
SPATH2: LODSB
        CMP     AL,AH         ;check one end
        JZ      SPATH5
;
        INC     SI
        CMP     AL,DL         ;check other end
        LOOPNZ  SPATH2
;
        JZ      SPATH5
;
; good prospect - put it in the flux list
;
SPATH3: XCHG    AX,DX
        STOSW
        MOV     DX,BP
        CMP     AH,DH
        JA      SPATH5        ;too far for best
;
; test for a new best path
;
        PUSH    BX
        CALL    TFLUX         ;get X,Y point
        PUSH    AX
        MOV     AX,0          ;ultimate dest
ULDEST  label   word
        CALL    CDIST         ;calculate dist
        POP     AX
        ADD     AH,DL
        JC      SPATH4
;
        CMP     AH,DH
        JBE     SPATHA        ;update best path
;
SPATH4: POP     BX
;
SPATH5: POP     SI
        POP     CX
;
SPATH6: LOOP    SPATH0
;
        MOV     CX,DI
        POP     SI
        SUB     CX,SI
        JZ      SPATH9        ;no flux this level
;
;- - - - - - - -
;
; test for next level flux
;
;   This loop has also been cleaned up. One
; improvement was to make the distances in the
; search path list cumulative. Before, I had a
; small loop prior to the main one that added up
; all the distances in the search path. Now, I
; just grab it with ADD AH,[DI-1]! What was I
; thinking?
;
;- - - - - - - -
;
        SHR     CX,1
        MOV     BX,BP
        MOV     BH,CH
        SHL     BX,1
        MOV     PFLPTR[BX],DI
        LEA     DI,OFS BUFFER+SPLIST[BX]
;
SPATH7: LODSW
        ADD     AH,[DI-1]     ;search path dist
        JC      SPATH8
;
        MOV     DX,BP
        MOV     DH,0          ;best distance
CBDIST  label   byte
        SUB     DH,AH
        JBE     SPATH8
;
        CALL    TFLUX         ;get X,Y point
        MOV     BP,DX
        PUSH    DI
        STOSW                 ;into search path
        PUSH    SI
        PUSH    CX
        CALL    SPATH         ;SPATH recursion
        POP     CX
        POP     SI
        POP     DI
;
SPATH8: LOOP    SPATH7
;
SPATH9: DEC     BP            ;prev search level
        RET
;
;- - - - - - - -
;
; found probable better path
;
;   The code here is a little register clumsy,
; but it is rarely executed. The registers are
; optimized for the loop above, where most of
; the time is spent.
;
;- - - - - - - -
;
SPATHA: MOV     BX,BP
        JB      SPATHB
;
; found equidistant path
;
;   Paths of equal distance are considered to be
; better if they have fewer flux. If desired,
; alternative paths could be identified here.
;   Another "best path" criteria could be travel
; through nebulae. Some shortest path travel
; courses take you through nebula, some don't.
; If the distances were the same, the path
; without nebula travel would be desirable.
;   Because there may be multiple "best paths",
; one might think that the answer for traveling
; from point A to point B may not always be the
; reverse of the answer for traveling from point
; B to point A. They always match, as the order
; in which the paths are tested is always the
; same, so long as the order of the flux list
; remains constant. The first occurance of a
; specific distance is the one stored.
;
        CMP     BL,0          ;best flux quantity
CBFLUX  label   byte
        JAE     SPATH4
;
; found better path
;
SPATHB: MOV     CBFLUX-1,BL   ;best flux quantity
        MOV     BH,AH
        MOV     BP,BX
        SUB     DH,AH
        SUB     CBDIST-1,DH   ;best distance
        MOV     CL,BL
        PUSH    DI
        MOV     DI,OFS BUFFER+CBPATH+2
        DEC     CX
        JZ      SPATHC
;
; end points from search path list
;
        MOV     SI,OFS BUFFER+SPLIST+2
        REP     MOVSW
;
; last end point from search loop
;
SPATHC: STOSW
        POP     DI
        JMP     SPATH4        ;back to loop
;
SPATH   ENDP
;
;-----------------------------------------------
;
CDLAX   PROC    NEAR
;
; Load X,Y point to AX, then calculate distance.
;
;Call With:
; SI = offset of X1, Y1
; BL = X2
; BH = Y2
;
;Returns:
; DL = distance
; SI = SI + 2
; all other registers preserved
;
        LODSW
;
;---------------
;
CDIST   PROC    NEAR
;
; Calculate distance between two X,Y points.
;
;   The distance returned represents the greater
; of |Dx| and |Dy|. The game does not penalize
; fuel consumption when traveling at angles.
; Although this fact complicates finding
; distance by map reconing, it makes the routine
; rather simple. I'm sure it simplified the game
; software, as there is no floating point math.
;
;Call With:
; AL = X1
; AH = Y1
; BL = X2
; BH = Y2
;
;Returns:
; DL = distance
; all other registers preserved
;
        PUSH    AX
        SUB     AL,BL
        JNC     CDIST0
;
        NEG     AL
;
CDIST0: SUB     AH,BH
        JNC     CDIST1
;
        NEG     AH
;
CDIST1: MOV     DL,AH
        CMP     AL,AH
        JB      CDIST2
;
        MOV     DL,AL
;
CDIST2: POP     AX
        RET
;
CDIST   ENDP
;
;---------------
;
CDLAX   ENDP
;
;-----------------------------------------------
;
TFLUX   PROC    NEAR
;
; Convert a flux termination number into an X,Y
; end point.
;
;Call With:
; AL = flux termination number (1 - 64)
;
;Returns:
; BL = flux termination X
; BH = flux termination Y
; all other registers preserved
;
        MOV     BX,2*FLXQTY
        SUB     BL,AL
        SHL     BX,1
        MOV     BX,WPT CFLUXS[BX]
        RET
;
TFLUX   ENDP
;
;-----------------------------------------------
;
FLOPP   PROC    NEAR
;
; Convert a flux termination number to the
; termination number for the opposite flux end.
;
;Call With:
; AL = flux termination number
;
;Returns:
; AL = opposite end flux termination number
; all other registers preserved
;
        TEST    AL,1
        JZ      FLOPP0
;
        INC     AX
        INC     AX
;
FLOPP0: DEC     AX
        RET
;
FLOPP   ENDP
;
;-----------------------------------------------
;
PARSE   PROC    NEAR
;
; Parse a numeric value from the command line.
;
;   Values greater than 255 or numeric-not-found
; errors set the carry flag on exit. The zero
; flag indicates the end of the command line.
;
;Call With:
; SI = current command line pointer
; CX = remaining command line length
;
;Returns:
; AL = numeric value
; SI, CX stepped for next parse
; CF = 0 if value valid
; ZF = 1 if end of command line
; BX, DI registers preserved
;
PARSE0: LODSB
        XOR     AL,30h
        CBW
        CMP     AL,DL
        JB      PARSE3
;
        LOOP    PARSE0
;
PARSE1: STC
        RET
;
PARSE2: LODSB
        XOR     AL,30h
        CMP     AL,DL
        JAE     PARSE4
;
        XCHG    AX,BP
        MUL     DL
        ADD     AX,BP
        OR      AH,AH
        JNZ     PARSE1
;
PARSE3: MOV     BP,AX
        LOOP    PARSE2
;
        CLC
        INC     CX
;
PARSE4: DEC     CX
        XCHG    AX,BP
        RET
;
PARSE   ENDP
;
;-----------------------------------------------
;
DSPLY   PROC    NEAR
;
; Display a line item.
;
;   Displays one travel or flux line item and
; sets the screen position to the start of the
; next line. Also, the point list source index
; and travel/flux flag are updated. Interrupt
; 21h, functions 2 and 9 are used so that output
; may be redirected. There are a number of ways
; to make this faster, speed is not critical.
;
;Call With:
; DL = 0 if travel item, 1 if flux item
; DH = 0Ah
; SI = offset into CBPATH point list
;
;Returns:
; DL = input DL value XOR 1
; SI = input SI value + 2
; DH register preserved
;
;Typical Output:
;  Travel: 125,100-126,87
;or
;    Flux: 126,87-173,88
;
        MOV     AX,OFS TRAVEL
        OR      DL,DL
        JZ      DSPLY0
;
        MOV     AX,OFS TRFLUX
;
DSPLY0: XCHG    AX,DX
        XOR     AL,1
        PUSH    AX
        DEC     AH
        PUSH    SI
        INT     21h
        POP     SI
        POP     DX
        LODSW
        CALL    DSPNT         ;display X,Y point
        MOV     AL,'-'
        CALL    DSCHR         ;display character
        MOV     AX,[SI]
        CALL    DSPNT         ;display X,Y point
;
;---------------
;
DCRLF   PROC    NEAR
;
; Display a carriage-return/line-feed sequence.
;
;Call With:
; nothing
;
;Returns:
; screen position set to start of next line
; BX, DX, SI registers preserved
;
        MOV     AL,0Dh
        CALL    DSCHR         ;display character
        MOV     AL,0Ah        ;divisor
        JMP SHO DSCHR0
;
DCRLF   ENDP
;
;---------------
;
DSPLY   ENDP
;
;-----------------------------------------------
;
DSPNT   PROC    NEAR
;
; Display one X,Y point.
;
;Call With:
; AL = X axis position
; AH = Y axis position
; DH = 0Ah
;
;Returns:
; point displayed at current screen position
; DX, SI registers preserved
;
        PUSH    AX
        CALL    DSVAL         ;display value
        MOV     AL,','
        CALL    DSCHR         ;display character
        POP     AX
        MOV     AL,AH
;
;---------------
;
DSVAL   PROC    NEAR
;
; Display one numeric value from 0 to 255.
;
;   Leading zeroes are supressed.
;
;Call With:
; AL = numeric value (0-255)
; DH = 0Ah
;
;Returns:
; value displayed at current screen position
; DX, SI registers preserved
;
        MOV     AH,0
        DIV     DH
        MOV     BL,AH
        CBW
        DIV     DH
        MOV     BH,AH
        OR      AL,AL
        JZ      DSVAL0
;
        CALL    DSDIG         ;display digit
        JMP SHO DSVAL1
;
DSVAL0: OR      BH,BH
        JZ      DSVAL2
;
DSVAL1: MOV     AL,BH
        CALL    DSDIG         ;display digit
;
DSVAL2: XCHG    AX,BX
;
;---------------
;
DSDIG   PROC    NEAR
;
; Display one numeric digit.
;
;Call With:
; AL = digit value (0-9)
;
;Returns:
; digit displayed at current screen position
; BX, DX, SI registers preserved
;
        OR      AL,30h
;
;---------------
;
DSCHR   PROC    NEAR
;
; Display one alpha-numeric character.
;
;Call With:
; AL = character
;
;Returns:
; character displayed at current screen position
; BX, DX, SI registers preserved
;
DSCHR0: XCHG    AX,DX
        PUSH    AX
        MOV     AH,2
        PUSH    BX
        PUSH    SI
        INT     21h
        POP     SI
        POP     BX
        POP     DX
        RET
;
DSCHR   ENDP
;
;---------------
;
DSDIG   ENDP
;
;---------------
;
DSVAL   ENDP
;
;---------------
;
DSPNT   ENDP
;
;-----------------------------------------------
;
; constant data
;
;---------------
;
TRAVEL  DB      '  Travel: ',24h
TRFLUX  DB      '    Flux: ',24h
TRDIST  DB      'Distance: ',24h
;
        EVEN
;
; continuum flux list (X1,Y1,X2,Y2)
;
;   The order of this list has a great effect on
; overall performance. At first look (that is,
; after I tried pre-search optimization), I
; tried ordering the flux according to the
; distance traversed (the greater of |Dx| and
; |Dy|, just like fuel consumption distance).
; I then realized that, in many paths, the
; opposite axis traverse was more important.
; I considered adding the two distances together
; to calculate order weight (the same as taking
; the average). That didn't sound right either.
; A flux is not better just because its slope is
; at an angle. The users origin and destination
; may be any two points on the map. The goal is
; overall performance, meaning the best order
; for all possible input combinations. I finally
; calculated the length of each line as if they
; were on a normal two dimensional plane. I
; believe this offers the best results. It's
; very impractical to test the theory, as there
; are so many possible input combinations.
;   Another good approach would be to move
; commonly interconnected flux toward the top of
; the list. For example, the flux near the Cross
; constellation are likely used in many paths,
; as they may be linked together in a number of
; different ways (Flux Grand Central). Some of
; the flux near the Spemim homeworld, on the
; other hand, aren't nearly as useful as they
; don't take you anywhere of intrest. Certainly,
; the 4 flux nearest to Starport are often used.
;
CFLUXS  DB      248,1,55,99     ;216.5
        DB      9,16,135,84     ;143.2
        DB      234,167,109,204 ;130.4
        DB      190,28,194,134  ;106.1
        DB      235,49,229,148  ;99.2
        DB      180,86,237,153  ;88.0
        DB      170,93,148,166  ;76.2
        DB      101,77,61,131   ;67.2
        DB      71,84,7,94      ;64.8
        DB      178,127,189,190 ;64.0
        DB      173,61,130,108  ;63.7
        DB      217,80,176,123  ;59.4
        DB      128,24,179,52   ;58.2
        DB      106,139,65,180  ;58.0
        DB      95,82,150,96    ;56.8
        DB      232,40,180,52   ;53.4
        DB      142,51,101,83   ;52.0
        DB      126,87,173,88   ;47.0
        DB      15,135,54,146   ;40.5
        DB      31,184,64,186   ;33.1
        DB      69,66,98,79     ;31.8
        DB      104,82,118,107  ;28.7
        DB      22,22,25,48     ;26.2
        DB      128,105,146,112 ;19.3
        DB      54,143,71,151   ;18.8
        DB      35,186,23,199   ;17.7
        DB      123,127,128,143 ;16.8
        DB      93,62,97,76     ;14.6
        DB      51,132,55,145   ;13.6
        DB      58,140,53,143   ;5.8
        DB      56,142,60,146   ;5.7
        DB      59,144,55,146   ;4.5
;
;---------------
;
; initialized variable data
;
;---------------
;
; prospective flux list pointers
; (initialized portion)
;
PFLPTR  DW      OFS BUFFER+PFLIST
;
;---------------
;
; uninitialized variable data
;
;---------------
;
BUFFER  LABEL   WORD
;
;   The following remarks indicate what the
; data would look like if created by defines.
; Referencing them via BUFFER eliminates several
; unnessecary zero bytes in the program file.
;   Although some of these data areas never get
; completely full, they are defined with the
; number of flux in mind. I know that no best
; path uses all 32 flux. While tracing execution
; in DEBUG, I've observed the routine testing a
; path with 15 flux. It probably doesn't go much
; deeper than that. I've seen a few best paths
; that use 7 flux. There are so many possible
; routes, calculation or measurement of actual
; use is impractical. The main concern is stack
; space, and 1316 bytes above the BUFFER label
; is of little consequence.
;
; N = total number of flux in game
;
; PFLPTR size = N words (uninitialized portion)
; CBPATH size = 2 x N + 2 words
; SPLIST size = N + 1 words
; PFLIST size = N x (N + 1) / 2 words
;
;- - - - - - - -
;
; uninitialized remainder of PFLPTR list
;
;       DW      32 DUP(?)
;
; current best path point list
;  low byte = X
; high byte = Y
; first X,Y point is always the users' origin
;
;CBPATH DW      66 DUP(?)
;
; search path list
; first word = 0 high byte for cumulative total
;   low byte = ending flux termination
;  high byte = distance consumed to termination
;
;SPLIST DW      33 DUP(?)
;
; prospective flux list(s)
;  low byte = ending flux termination
; high byte = distance from previous origin
;
;PFLIST DW      528 DUP(?)
;
;---------------
;
; display strings
; (overwritten by data above BUFFER)
;
MTITLE  DB      0Dh,0Ah,'STARFLIGHT I Shortest'
        DB      ' Path Finder V 1.1',0Dh,0Ah
        DB      '(C)opyright 1989-2000, '
        DB      'David R. Sheldon',0Dh,0Ah
;
MUSAGE  DB      0Dh,0Ah,'Use: SF1SPF11 Xo,Yo'
        DB      ' Xd,Yd',0Dh,0Ah,0Dh,0Ah
        DB      ' Xo = origin X',0Dh,0Ah
        DB      ' Yo = origin Y',0Dh,0Ah
        DB      ' Xd = destination X',0Dh,0Ah
        DB      ' Yd = destination Y',0Dh,0Ah
        DB      0Dh,0Ah,'Notes:',0Dh,0Ah
        DB      '1) valid X values range'
        DB      ' from 0 to 249',0Dh,0Ah
        DB      '2) valid Y values range'
        DB      ' from 0 to 219',0Dh,0Ah
        DB      '3) no penalty for 45 '
        DB      'degree travel',0Dh,0Ah
        DB      '4) non-numeric characters'
        DB      ' are skipped',0Dh,0Ah,'   '
        DB      '(commas are optional)',0Dh,0Ah
        DB      '5) input beyond 4 values'
        DB      ' is ignored',0Dh,0Ah,24h
;
;-----------------------------------------------
;
_TEXT   ENDS
;
        END     _main
;
